#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <utility.h>

#include "lf.h"
#include "soggetti.h"
#include "sezioni.h"
#include "convoc.h"
#include "rconvoc.h"

#include "at4.h"
#include "at4400a.h"
#include "at4100b.h"
#include "at4100c.h"

#define ALIAS_LCP	100
#define ALIAS_TCS	200

enum ts { undefined = 0, elenco = 1, etichette = 2, cartoline = 3 };

// definizione form per etichette
class TStConvoc_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TStConvoc_form(): TForm() {};
  TStConvoc_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TStConvoc_form() {};
};

class TStConvoc_app : public TPrintapp
{
  TRelation*				_rel;
  TMask*      			_msk;
  TStConvoc_form*		_form_eti;
  TStConvoc_form*   _form_car;
	  
  int 							_cur1;
  TParagraph_string _cognome_nome;
  TDate 						_data_stampa;
  TDate							_dataini, _datafin;
  bool 							_aggiorna;
	TString16 				_punto,_tipo, _sezione, _sottog, _ordina;
	TString16					_codsez, _codsot;
	TString16					_orario, _datacart;
	TString80					_invitoper, _data, _presso1, _presso2, _presso3, _presso4;
	TString80					_note, _intest1, _intest2, _intest3, _intest4;
	bool 							_usasez;
	ts								_tipo_stampa;

  //static bool filter_func_stconvoc(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  ts dati_cartoline();

public:
  void crea_intestazione();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void dati_sezione(const TString16 codsez, const TString16 codsot);
  TMask& app_mask() { return *_msk; }
  TStConvoc_app() : _data_stampa(TODAY), _cognome_nome("",25) {}
};

HIDDEN inline TStConvoc_app& app() { return (TStConvoc_app&) main_app(); }

TCursor* TStConvoc_form::cursor() const { return app().current_cursor(); }

TRelation* TStConvoc_form::relation() const { return cursor()->relation(); }

ts TStConvoc_app::dati_cartoline()
{
	TMask msk("at4100b");
 	if (msk.run() == K_ENTER) 
 	{ 
 		_invitoper = msk.get(F_INVITOPER);
 		_data			 = msk.get(F_DATA);
 		_presso1   = msk.get(F_PRESSO1);
 		_presso2   = msk.get(F_PRESSO2);
 		_presso3   = msk.get(F_PRESSO3);
 		_presso4   = msk.get(F_PRESSO4);
 		_datacart  = msk.get(F_DATACART);
 		_orario  	 = msk.get(F_ORARIO);
 		_note   	 = msk.get(F_NOTE);
 		_usasez		 = msk.get_bool(F_USASEZ);
 		return cartoline;
 	}
 	else
 		return undefined;
}

void TStConvoc_app::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipo_stampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			corpo.update();
  		for (int i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case cartoline:
		{
			TPrint_section& corpo = _form_car->get_body();
			const TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
			const TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
			if ((_codsez!=codsez)||(_codsot!=codsot))
			{
				_codsez = codsez;
				_codsot = codsot;
				dati_sezione(codsez, codsot);
				corpo.reset();
			  TForm_item& intest1 = corpo.find_field(CAR_INTEST1);
			  intest1.set(_intest1);
			  TForm_item& intest2 = corpo.find_field(CAR_INTEST2);
			  intest2.set(_intest2);
			  TForm_item& intest3 = corpo.find_field(CAR_INTEST3);
			  intest3.set(_intest3);
			  TForm_item& intest4 = corpo.find_field(CAR_INTEST4);
			  intest4.set(_intest4);
			  TForm_item& invito = corpo.find_field(CAR_INVITOPER);
	  		invito.set(_invitoper);
			  TForm_item& data = corpo.find_field(CAR_DATA);
	  		data.set(_data);
			  TForm_item& presso1 = corpo.find_field(CAR_PRESSO1);
	  		presso1.set(_presso1);
			  TForm_item& presso2 = corpo.find_field(CAR_PRESSO2);
	  		presso2.set(_presso2);
			  TForm_item& presso3 = corpo.find_field(CAR_PRESSO3);
	  		presso3.set(_presso3);
			  TForm_item& presso4 = corpo.find_field(CAR_PRESSO4);
	  		presso4.set(_presso4);
			  TForm_item& orario = corpo.find_field(CAR_ORARIO);
	  		orario.set(_orario);
			  TForm_item& datacart = corpo.find_field(CAR_DATACART);
	  		datacart.set(_datacart);
			  TForm_item& note = corpo.find_field(CAR_NOTE);
	  		note.set(_note);
			}
			corpo.update();
  		for (int i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"#########"));
			set_row(1,"@10g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@13g@S", FLD(LF_SOGGETTI,SOG_COGNOME));
			set_row(1,"@39g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@50g@ld", FLD(LF_SOGGETTI,SOG_DATAPROSSI));
			set_row(1,"@61g@ld", FLD(LF_SOGGETTI,SOG_DATAPROSAF));
			set_row(1,"@72g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(1,"@83g@ld", FLD(LF_SOGGETTI,SOG_DATAULTID));
			set_row(1,"@94g@S", FLD(LF_SOGGETTI,SOG_IDON1));
			set_row(1,"@97g@S", FLD(LF_SOGGETTI,SOG_IDON2));
			set_row(1,"@100g@S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(1,"@105g@ld", FLD(LF_RCONVOC,RCV_DATACONV));

			set_row(2,"@13g@S", FLD(LF_SOGGETTI,SOG_NOME));
			set_row(2,"@53g@pn", FLD(LF_SOGGETTI,SOG_INTSI,"###"));
			set_row(2,"@64g@pn", FLD(LF_SOGGETTI,SOG_INTAF,"###"));
			set_row(2,"@76g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(2,"@87g@S", FLD(LF_SOGGETTI,SOG_TIPOULTID));
			set_row(2,"@94g@S", FLD(LF_SOGGETTI,SOG_IDON3));
			set_row(2,"@97g@S", FLD(LF_SOGGETTI,SOG_IDON4));
			set_row(2,"@100g@S", FLD(LF_SOGGETTI,SOG_RHANTID));			
			set_row(2,"@105g@S", FLD(LF_RCONVOC,RCV_CHIAMATA));
			
			set_row(3, "");			
		}
		break;
	}					
}

bool TStConvoc_app::preprocess_page(int file, int counter)
{
	// contatore soggetti stampati 
	// per ora non c'

	switch (_tipo_stampa)
	{
	case elenco:
	{
		TString80 nome = current_cursor()->curr(LF_SOGGETTI).get(SOG_COGNOME);
		nome << " ";
		nome << current_cursor()->curr(LF_SOGGETTI).get(SOG_NOME);
		_cognome_nome = nome;

		// salto pagina se cambio sezione
		const TString16 codsez = current_cursor()->curr(LF_SOGGETTI).get(SOG_CODSEZ);		
		const TString16 codsot = current_cursor()->curr(LF_SOGGETTI).get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				printer().formfeed();
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}
	break;
	}
  return TRUE;
}

void TStConvoc_app::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if ((densot.ok())&& (densot.not_empty()))
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

void TStConvoc_app::dati_sezione(const TString16 codsez, const TString16 codsot)
{
	_intest1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ1);
	_intest2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ2);
	_intest3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ3);
	_intest4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ4);
	if (_usasez)
	{
		if (_presso1.blank())
		{
			_presso1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE1);
			_presso2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE2);
			_presso3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE3);
			_presso4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE4);
		}			
		if (_orario.blank())
			_orario = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVORA);
		if (_note.blank())
			_note = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVNOTE);
	}
	return;
}

bool TStConvoc_app::set_print(int m)
{ 
	_tipo_stampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipo_stampa = elenco;
  break;
  case F_ETICHETTE:
  	_tipo_stampa = etichette;	
	break;  	
  case F_CARTOLINE:
  	_tipo_stampa = dati_cartoline();	
  break;
  }	
	if (_tipo_stampa != undefined)
  {
    _dataini 		= _msk->get(F_DATAINI);
    _datafin 		= _msk->get(F_DATAFIN);
    _punto      = _msk->get(F_PUNTO);
    _tipo				= _msk->get(F_TIPO);
    _codsez     = _msk->get(F_CODSEZ);
    _codsot			= _msk->get(F_CODSOT);
    _aggiorna   = _msk->get_bool(F_AGGIORNA);
    _ordina     = _msk->get(F_ORDINA);
		TString256 chiave = "";
		switch (_ordina[0])
		{
			case 'A':
				chiave = "90->CODSEZ|90->CODSOT|UPPER(90->COGNOME)|UPPER(90->NOME)|116->DATACONV";
			break;
			case 'C':
				chiave = "90->CODSEZ|90->CODSOT|90->DOM_CAP|UPPER(90->COGNOME)|UPPER(90->NOME)|116->DATACONV";
			break;
		}
		TString80 filtro = "";			
  	// filtro per punto di raccolta
  	if (_punto.not_empty())
  	{
  		if (filtro.not_empty())
				filtro << " && ";
			filtro << format("(115->PUNTO == \"%s\")",(const char*)_punto);
		}	
		// filtro per tipo
		if (_tipo.not_empty())
		{
  		if (filtro.not_empty())
				filtro << " && ";
			filtro << format("(115->TIPO == \"%s\")",(const char*)_tipo);
		}			
		// filtro per sezione/sottogruppo
		if (_sezione.not_empty())
		{
  		if (filtro.not_empty())
				filtro << " && ";
 			filtro << format("(90->CODSEZ == \"%s\")",(const char*)_sezione);
			if (_sottog.not_empty())
			{
	  		if (filtro.not_empty())
					filtro  << " && ";
				filtro << format("(90->CODSOT == \"%s\")",(const char*)_sottog);
			}
		}
		// filtro per data
	  TRectype da(LF_RCONVOC);
  	TRectype a (LF_RCONVOC);
  	if (_dataini.ok())
			da.put(RCV_DATACONV, _dataini);
  	if (_datafin.ok())
			a.put(RCV_DATACONV, _datafin);
	  _cur1 = add_cursor(new TSorted_cursor(_rel, (const char*) chiave,"", 2, &da, &a));
		current_cursor()->setfilter((const char*) filtro, TRUE);
		//current_cursor()->set_filterfunction(filter_func_stconvoc);
  	_codsez = "**";
  	_codsot = "**";
    reset_files(); 
    add_file(LF_RCONVOC);
		reset_print();
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStConvoc_app::crea_intestazione()
{
  reset_header();
  
  if (_tipo_stampa == elenco)
  {
  	TString sep(132);
  	sep = "STAMPA CONVOCATI";
  	if (_tipo.not_empty())
  	{
  		sep << " per ";
  		sep << _tipo;
		}  		
  	sep << " dal ";
 		sep << _dataini.string();
 		sep << " al ";
 		sep << _datafin.string();
 		sep << " - PUNTO ";
 		sep << _punto;
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);

  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@120g%s", (const char*) sep);
  	
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);

	  set_header(5,"@0gCod.@10gC.@13gCognome@39gNato il@50gData pr.SI@61gData pr.AF@72gData/Tipo@83gData/Tipo@94gIdon.@100gAB0@105gData conv.");
	  set_header(6,"@13gNome@50gInterv. SI@61gInterv. AF@72gult. don.@83gult. idon.@94gper@100gRh@105gTipo conv.");  
	  set_header(7,"@0g---------@10g--@13g-------------------------@39g----------@50g----------@61g----------@72g----------@83g----------@94g-----@100g----@105g----------");
	}  	
}

bool TStConvoc_app::user_create()
{
  _rel = new TRelation(LF_RCONVOC);
  _rel->add(LF_SOGGETTI, "CODICE==CODICE");
  _rel->add(LF_CONVOC, "NUMERO==NUMERO");
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,LF_SOGGETTI,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,LF_SOGGETTI);
  // per stampare la denominazione della sezione nell'intestazione
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT",1,LF_SOGGETTI);
  _msk = new TMask("at4400a");
  _form_eti = new TStConvoc_form("AT_ETSOG");
  _form_car = new TStConvoc_form("AT_ETCAR");

  return TRUE;
}

bool TStConvoc_app::user_destroy()
{
  delete _msk;
  delete _rel;
  delete _form_eti;
  delete _form_car;
  return TRUE;
}

int at4400(int argc, char* argv[])
{

  TStConvoc_app a;

  a.run(argc, argv, "Stampa convocazioni Bologna");

  return 0;
}
