#include <applicat.h>
#include <config.h>
#include <mask.h>   
#include <printer.h>
#include <relation.h>
#include <tabutil.h>
#include <urldefid.h>
#include <utility.h>
#include <form.h>
#include <printapp.h>

#include "..\ve\veconf.h"
#include "mglib.h"
#include "mg3100.h"
#include "mg3frm.h"


class TForm_giacprezzi : public TForm_stampemg
{
  TConfig *conf_ditta;
  TCondizione_vendita * listino;
public:
  void set_codlistino(const char * codcm,const char * cod);
  bool validate(TForm_item &cf, TToken_string &s);
  TForm_giacprezzi(const char *name,const char *code) ;
  ~TForm_giacprezzi();
};

bool TForm_giacprezzi::validate(TForm_item &cf, TToken_string &s)
{ 
  const TString code(s.get(0)); // prende il primo parametro, il codice del messaggio
  TString subcode;

  if (code=="_USER") {
    subcode=s.get();
    if (subcode=="PREZZO") {
      TString codart(relation()->lfile(LF_ANAMAG).get("CODART"));
      if (listino->ricerca(codart,0.0))
        cf.set(listino->get_prezzo().string());
      return TRUE;
    }
  }
  return TForm_stampemg::validate(cf, s);
}

void TForm_giacprezzi::set_codlistino(const char * codcatv,const char * codlist)
{
  listino->put_listino(codlist,codcatv);
}

TForm_giacprezzi::TForm_giacprezzi(const char *name,const char *code) :
  TForm_stampemg(name,code)
{
  listino=new TCondizione_vendita(new TConfig(CONFIG_DITTA,""));
  listino->set_anamag(relation()->lfile(LF_ANAMAG));
  listino->set_umart(relation()->lfile(LF_UMART));
}

TForm_giacprezzi::~TForm_giacprezzi()
{
  delete &(listino->config_ditta());
  delete listino;
}



class TStampadatist_mask: public TMask
{
  static bool handle_ragg(TMask_field &fld, KEY k); // handler 
  static bool handle_liv(TMask_field &fld, KEY k); // handler 
  static bool handle_subord(TMask_field &fld, KEY k); // handler 
public:
  TStampadatist_mask();
  ~TStampadatist_mask(){};
};

TStampadatist_mask::TStampadatist_mask():
  TMask("mg3100")
{
  TConfig prassid(CONFIG_DITTA, "ve"); // apre il file di configurazione della ditta corrente
  if (prassid.get_bool("GES", NULL, A_LISTINI))
    field(F_CATVENLISTINO).enable(prassid.get_bool("GESLISCV"));
  else {
    field(F_CATVENLISTINO).disable(); field(F_CODLISTINO).disable();
  }
  TCodgiac_livelli giaclev;
  TCodart_livelli artlev;
  
  set_handler(F_TOLIVELLOART, handle_liv);
  set_handler(F_TOLIVELLOGIAC, handle_liv);
  set_handler(F_ORDINE, handle_ragg);
  set_handler(F_DETTAGLIODEP, handle_ragg);
  set_handler(F_DETTAGLIOMAG, handle_ragg);
  set_handler(F_ORDINEART,handle_subord);
  if (!artlev.enabled()) {
    TMask_field &f1=field(F_RAGGCODICE);
    f1.reset(); f1.check();f1.hide();
  }
  if (!giaclev.enabled()) {
    TMask_field &f2=field(F_RAGGLIVGIAC);
    f2.reset(); f2.check();f2.hide();
  }
}

bool TStampadatist_mask::handle_subord(TMask_field &fld, KEY k)
{
  if (k == K_SPACE) 
  {
    TOperable_field &fraggcod=(TOperable_field &)fld.mask().field(F_RAGGCODICE);
    TOperable_field &fragggiac=(TOperable_field &)fld.mask().field(F_RAGGLIVGIAC);
    switch (fld.get()[1]) {
      case 'C':
        if (fld.mask().get(F_ORDINE)=="A") 
          fraggcod.enable();
        else {
          if (!fld.mask().get_bool(F_DETTAGLIOMAG) || !fld.mask().get_bool(F_DETTAGLIODEP))
          {
            fraggcod.reset(); fraggcod.disable();
            fragggiac.reset(); fragggiac.disable();
            fraggcod.do_message(0);fragggiac.do_message(0);
          }
        }
      break;
      case 'M':
      case 'D':
        fraggcod.reset(); fraggcod.check();fraggcod.disable();
      break;
    }
  }
  return TRUE;
}

bool TStampadatist_mask::handle_liv(TMask_field &fld, KEY k)
{
  if (k == K_TAB) 
  {
    TOperable_field &ftotmag=(TOperable_field &)fld.mask().field(F_TOTALIMAGAZZINI);
    TOperable_field &fdettmag=(TOperable_field &)fld.mask().field(F_DETTAGLIOMAG);
    if (fld.mask().get(F_ORDINE)=="A") 
    {
      if (fld.mask().get_int(F_TOLIVELLOART) || fld.mask().get_int(F_TOLIVELLOGIAC))
      {
        // raggruppamenti prima del dettaglio magazzini
        ftotmag.reset(); ftotmag.disable();ftotmag.do_message(0);
      } else {
        ftotmag.enable();
      }
    }
  }
  return TRUE;
}

bool TStampadatist_mask::handle_ragg(TMask_field &fld, KEY k)
{
  if (k == K_SPACE) 
  {
    TOperable_field &ftotmag=(TOperable_field &)fld.mask().field(F_TOTALIMAGAZZINI);
    TOperable_field &fraggcod=(TOperable_field &)fld.mask().field(F_RAGGCODICE);
    TOperable_field &fragggiac=(TOperable_field &)fld.mask().field(F_RAGGLIVGIAC);
    if (fld.mask().get(F_ORDINE)=="M")
    {
      ftotmag.enable();
      if (!fld.mask().get_bool(F_DETTAGLIOMAG) || !fld.mask().get_bool(F_DETTAGLIODEP))
      {
        fraggcod.reset(); fraggcod.disable();
        fragggiac.reset(); fragggiac.disable();
        fraggcod.do_message(0);fragggiac.do_message(0);
        return TRUE;
      }
    } 
    fraggcod.enable();
    fragggiac.enable();
  }
  return TRUE;
}


// STAMPA anagrafiche
class TStampa_datistorici : public TPrint_application
{
  TArray * _files;
  TMask * _mask;
  TCursor * _cur;
  TForm_giacprezzi * _form; // to be moved into TPrint_application

protected:
  virtual bool print_one(int file) 
  {return TRUE;}
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int i=1);
  virtual void set_page(int file,int count)
  { }
  void setprint_permagazzini();
  void setprint_perarticoli();

public:
  TStampa_datistorici() {}
};


bool TStampa_datistorici::user_create()
{
  _mask = new TStampadatist_mask();
  _files = new TArray();
  _files->add(new TLocalisamfile(LF_MAG));
  return TRUE;
}

bool TStampa_datistorici::user_destroy()
{
  delete _mask;
  delete _files;
  return TRUE;
}

void TStampa_datistorici::setprint_permagazzini()
{
  char subordine=*_mask->get(F_ORDINEART);
  TRectype darec(LF_MAG),arec(LF_MAG);
  TString cfilter,sortexp,joinexp;
  _form = new TForm_giacprezzi("mg3100b", "");
  _cur = _form->cursor();
  // setta il filtro sul cursore
  if (*_mask->get(F_DAART))
    cfilter << "(CODART>=" <<'"' << _mask->get(F_DAART)<< "\")&&" ;
  if (*_mask->get(F_AART))
    cfilter << "(CODART<=" <<'"' << _mask->get(F_AART)<< "\")&&" ;
  switch (*_mask->get(F_FILTRO)) {
    case '0': // non nulli
      cfilter << "(STR(" << LF_MAG << "->GIAC!=\"0\"))&&";
    break;
    case 'P': // positivi
      cfilter << "(STR(" << LF_MAG << "->GIAC>\"0\"))&&";
    break;
    case 'N': // negativi
      cfilter << "(STR(" << LF_MAG << "->GIAC<\"0\"))&&";
    break;
    case 'S': // sottoscorta
      cfilter << "(STR(GIAC<SCORTAMIN))&&";
    break;
  }
  if (cfilter.not_empty())
    cfilter.cut(cfilter.len()-2); 

  // abilita/disabilita le sezioni dei totali
  _form->setdett_permag(_mask->get_int(F_FROMLIVELLOART),
    _mask->get_int(F_TOLIVELLOART),
    _mask->get_int(F_FROMLIVELLOGIAC),
    _mask->get_int(F_TOLIVELLOGIAC),
    TRUE,
    _mask->get_bool(F_TOTALIDEPOSITI),
    _mask->get_bool(F_DETTAGLIODEP));

  _form->find_field('H',odd_page,FF_FLAGGIAC).set((_mask->get_bool(F_STAMPAGIAC) && *_mask->get(F_FILTRO)!='S') ? "1": "0");
  _form->find_field('H',odd_page,FF_FLAGSOTTOS).set((_mask->get_bool(F_STAMPAGIAC) && *_mask->get(F_FILTRO)=='S') ? "1": "0");
  _form->find_field('B',odd_page,FF_FLAGVALMEDI).set(_mask->get_bool(F_VALMEDI) ? "1": "0");

  _form->find_field('H',odd_page,FF_PREZZIVALORI).set(_mask->get(F_PREZZIVALORI));
  _form->find_field('B',odd_page,FF_PREZZIVALORI).set(_mask->get(F_PREZZIVALORI));
  _form->set_codlistino(_mask->get(F_CATVENLISTINO),_mask->get(F_CODLISTINO));

  darec.put("ANNOES",_mask->get(F_ANNOES));
  darec.put("CODMAG",_mask->get(F_DAMAG));
  arec.put("ANNOES",_mask->get(F_ANNOES));
  arec.put("CODMAG",_mask->get(F_AMAG));
  _cur->setfilter(cfilter);
  _cur->setregion(darec,arec);
}

// ORDINAMENTO 
void TStampa_datistorici::setprint_perarticoli()
{
  TRectype darec(LF_ANAMAG),arec(LF_ANAMAG);
  TString cfilter,filter,joinexp;
  TString sortexp("ANNOES|CODART|LIVELLO|CODMAG");

  char subordine=*_mask->get(F_ORDINEART);
  _form = new TForm_giacprezzi("mg3100a", "");
  _cur = _form->cursor();
  if (*_mask->get(F_DAMAG))
    filter << "(CODMAG[1,3]>=" <<'"' << _mask->get(F_DAMAG)<< "\")&&" ;
  if (*_mask->get(F_AMAG))
    filter << "(CODMAG[1,3]<=" <<'"' << _mask->get(F_AMAG)<< "\")&&" ;
  switch (*_mask->get(F_FILTRO)) {
    case '0': // non nulli
      filter << "(STR(" << LF_MAG << "->GIAC!=\"0\"))&&";
    break;
    case 'P': // positivi
      filter << "(STR(" << LF_MAG << "->GIAC>\"0\"))&&";
    break;
    case 'N': // negativi
      filter << "(STR(" << LF_MAG << "->GIAC<\"0\"))&&";
    break;
    case 'S': // sottoscorta
      filter << "(STR(GIAC<SCORTAMIN))&&";
    break;
  }
  if (filter.not_empty())
    filter.cut(filter.len()-2); 
  switch (subordine)
  {
    case 'C':
      _cur->setkey(1);
      darec.put("CODART",_mask->get(F_DAART));
      arec.put("CODART",_mask->get(F_AART));
      _form->find_field('B',odd_page,"H_CATMER").hide();
      _form->find_field('B',odd_page,"TOT_CATMER").hide();
      _form->setcatmer(FALSE);
    break;
    case 'D':
      _cur->setkey(2);
      darec.put("DESCR",_mask->get(F_DADES));
      arec.put("DESCR",_mask->get(F_ADES));
      _form->find_field('B',odd_page,"H_CATMER").disable();
      _form->find_field('B',odd_page,"TOT_CATMER").hide();
      _form->setcatmer(FALSE);
    break;
    case 'M':
      _cur->setkey(3);
      darec.put("GRMERC",_mask->get(F_DACATMER));
      darec.put("CODART",_mask->get(F_DAART));
      arec.put("GRMERC",_mask->get(F_ACATMER));
      arec.put("CODART",_mask->get(F_AART));
      _form->find_field('B',odd_page,"H_CATMER").show();
      _form->find_field('B',odd_page,"TOT_CATMER").show();
      _form->setcatmer(TRUE);
    break;
  }
    _form->setdett_perart(
      _mask->get_int(F_FROMLIVELLOART),
      _mask->get_int(F_TOLIVELLOART),
      _mask->get_int(F_FROMLIVELLOGIAC),
      _mask->get_int(F_TOLIVELLOGIAC),
      _mask->get_bool(F_TOTALIMAGAZZINI),
      _mask->get_bool(F_TOTALIDEPOSITI));

  _form->find_field('H',odd_page,FF_FLAGGIAC).set((_mask->get_bool(F_STAMPAGIAC) && *_mask->get(F_FILTRO)!='S') ? "1": "0");
  _form->find_field('H',odd_page,FF_FLAGSOTTOS).set((_mask->get_bool(F_STAMPAGIAC) && *_mask->get(F_FILTRO)=='S') ? "1": "0");
  _form->find_field('B',odd_page,FF_FLAGVALMEDI).set(_mask->get_bool(F_VALMEDI) ? "1": "0");

  _form->find_field('H',odd_page,FF_PREZZIVALORI).set(_mask->get(F_PREZZIVALORI));
  _form->find_field('B',odd_page,FF_PREZZIVALORI).set(_mask->get(F_PREZZIVALORI));
  _form->set_codlistino(_mask->get(F_CATVENLISTINO),_mask->get(F_CODLISTINO));

  joinexp << "ANNOES==" << _mask->get(F_ANNOES) << "|CODART==CODART";

  TSortedfile *mag;
  mag= new TSortedfile(LF_MAG,NULL,sortexp,filter,1);
  _cur->relation()->replace(mag,1,joinexp);
  _cur->setfilter(cfilter);
  _cur->setregion(darec,arec);
}


bool TStampa_datistorici::set_print(int i)
{
  while (TRUE) {
    if (_mask->run() != K_ENTER) 
      return FALSE;

    if (*_mask->get(F_ORDINE)=='A')  
    {
      setprint_perarticoli();
    } else {                      
      setprint_permagazzini();
    }
    _form->print();
    delete _form;
  } // while true  
  return TRUE;
}


int mg3100(int argc, char* argv[])
{
  TStampa_datistorici a;// derivata da Application e con uso di form
  a.run(argc,argv,"Stampa dati storici di giacenza");
  return 0;
}
